@@ -1039,14 +1039,8 @@ pub(crate) fn expression(expr: &Expr, checker: &Checker) {
                 flake8_simplify::rules::zip_dict_keys_and_values(checker, call);
             }
             if checker.any_rule_enabled(&[
-                Rule::OsChmod,
-                Rule::OsMkdir,
-                Rule::OsMakedirs,
-                Rule::OsRename,
-                Rule::OsReplace,
                 Rule::OsStat,
                 Rule::OsPathJoin,
-                Rule::OsPathSamefile,
                 Rule::OsPathSplitext,
                 Rule::BuiltinOpen,
                 Rule::PyPath,
@@ -1112,6 +1106,24 @@ pub(crate) fn expression(expr: &Expr, checker: &Checker) {
                 if checker.is_rule_enabled(Rule::OsGetcwd) {
                     flake8_use_pathlib::rules::os_getcwd(checker, call, segments);
                 }
+                if checker.is_rule_enabled(Rule::OsChmod) {
+                    flake8_use_pathlib::rules::os_chmod(checker, call, segments);
+                }
+                if checker.is_rule_enabled(Rule::OsRename) {
+                    flake8_use_pathlib::rules::os_rename(checker, call, segments);
+                }
+                if checker.is_rule_enabled(Rule::OsReplace) {
+                    flake8_use_pathlib::rules::os_replace(checker, call, segments);
+                }
+                if checker.is_rule_enabled(Rule::OsPathSamefile) {
+                    flake8_use_pathlib::rules::os_path_samefile(checker, call, segments);
+                }
+                if checker.is_rule_enabled(Rule::OsMkdir) {
+                    flake8_use_pathlib::rules::os_mkdir(checker, call, segments);
+                }
+                if checker.is_rule_enabled(Rule::OsMakedirs) {
+                    flake8_use_pathlib::rules::os_makedirs(checker, call, segments);
+                }
                 if checker.is_rule_enabled(Rule::PathConstructorCurrentDirectory) {
                     flake8_use_pathlib::rules::path_constructor_current_directory(
                         checker, call, segments,
