@@ -1,4 +1,5 @@
 //! Rules from [flake8-pytest-style](https://pypi.org/project/flake8-pytest-style/).
+mod helpers;
 pub(crate) mod rules;
 pub mod settings;
 pub mod types;
@@ -11,9 +12,9 @@ mod tests {
     use test_case::test_case;
 
     use crate::registry::Rule;
-    use crate::settings::types::{IdentifierPattern, PreviewMode};
+    use crate::settings::types::IdentifierPattern;
     use crate::test::test_path;
-    use crate::{assert_messages, settings};
+    use crate::{assert_diagnostics, settings};
 
     use super::settings::Settings;
     use super::types;
@@ -354,37 +355,7 @@ mod tests {
                 ..settings::LinterSettings::for_rule(rule_code)
             },
         )?;
-        assert_messages!(name, diagnostics);
-        Ok(())
-    }
-
-    #[test_case(
-        Rule::PytestRaisesWithMultipleStatements,
-        Path::new("PT012.py"),
-        Settings::default(),
-        "PT012_preview"
-    )]
-    #[test_case(
-        Rule::PytestWarnsWithMultipleStatements,
-        Path::new("PT031.py"),
-        Settings::default(),
-        "PT031_preview"
-    )]
-    fn test_pytest_style_preview(
-        rule_code: Rule,
-        path: &Path,
-        plugin_settings: Settings,
-        name: &str,
-    ) -> Result<()> {
-        let diagnostics = test_path(
-            Path::new("flake8_pytest_style").join(path).as_path(),
-            &settings::LinterSettings {
-                preview: PreviewMode::Enabled,
-                flake8_pytest_style: plugin_settings,
-                ..settings::LinterSettings::for_rule(rule_code)
-            },
-        )?;
-        assert_messages!(name, diagnostics);
+        assert_diagnostics!(name, diagnostics);
         Ok(())
     }
 
@@ -403,7 +374,7 @@ mod tests {
                 ])
             },
         )?;
-        assert_messages!("PT006_and_PT007", diagnostics);
+        assert_diagnostics!("PT006_and_PT007", diagnostics);
         Ok(())
     }
 }
