@@ -1,11 +1,11 @@
 use std::fmt;
 
-use ruff_diagnostics::{Diagnostic, Violation};
 use ruff_macros::{ViolationMetadata, derive_message_formats};
 use ruff_python_semantic::{Binding, BindingKind};
 use ruff_text_size::Ranged;
 
-use crate::Locator;
+use crate::Violation;
+use crate::checkers::ast::Checker;
 
 /// ## What it does
 /// Checks for the use of non-ASCII characters in variable names.
@@ -44,10 +44,11 @@ impl Violation for NonAsciiName {
 }
 
 /// PLC2401
-pub(crate) fn non_ascii_name(binding: &Binding, locator: &Locator) -> Option<Diagnostic> {
+pub(crate) fn non_ascii_name(checker: &Checker, binding: &Binding) {
+    let locator = checker.locator();
     let name = binding.name(locator.contents());
     if name.is_ascii() {
-        return None;
+        return;
     }
 
     let kind = match binding.kind {
@@ -73,17 +74,17 @@ pub(crate) fn non_ascii_name(binding: &Binding, locator: &Locator) -> Option<Dia
         | BindingKind::Deletion
         | BindingKind::ConditionalDeletion(_)
         | BindingKind::UnboundException(_) => {
-            return None;
+            return;
         }
     };
 
-    Some(Diagnostic::new(
+    checker.report_diagnostic(
         NonAsciiName {
             name: name.to_string(),
             kind,
         },
         binding.range(),
-    ))
+    );
 }
 
 #[derive(Debug, PartialEq, Eq, Copy, Clone)]
