@@ -7,16 +7,18 @@ pub(crate) mod types;
 
 #[cfg(test)]
 mod tests {
+    use std::collections::BTreeSet;
     use std::path::Path;
 
     use anyhow::Result;
     use ruff_python_ast::PythonVersion;
+    use ruff_python_semantic::{MemberNameImport, NameImport};
     use test_case::test_case;
 
     use crate::registry::Rule;
-    use crate::rules::pyupgrade;
+    use crate::rules::{isort, pyupgrade};
     use crate::settings::types::PreviewMode;
-    use crate::test::test_path;
+    use crate::test::{test_path, test_snippet};
     use crate::{assert_diagnostics, settings};
 
     #[test_case(Rule::ConvertNamedTupleFunctionalToClass, Path::new("UP014.py"))]
@@ -99,7 +101,8 @@ mod tests {
     #[test_case(Rule::UnnecessaryClassParentheses, Path::new("UP039.py"))]
     #[test_case(Rule::UnnecessaryDefaultTypeArgs, Path::new("UP043.py"))]
     #[test_case(Rule::UnnecessaryEncodeUTF8, Path::new("UP012.py"))]
-    #[test_case(Rule::UnnecessaryFutureImport, Path::new("UP010.py"))]
+    #[test_case(Rule::UnnecessaryFutureImport, Path::new("UP010_0.py"))]
+    #[test_case(Rule::UnnecessaryFutureImport, Path::new("UP010_1.py"))]
     #[test_case(Rule::UselessMetaclassType, Path::new("UP001.py"))]
     #[test_case(Rule::UselessObjectInheritance, Path::new("UP004.py"))]
     #[test_case(Rule::YieldInForLoop, Path::new("UP028_0.py"))]
@@ -294,4 +297,70 @@ mod tests {
         assert_diagnostics!(diagnostics);
         Ok(())
     }
+
+    #[test]
+    fn i002_conflict() {
+        let diagnostics = test_snippet(
+            "from pipes import quote, Template",
+            &settings::LinterSettings {
+                isort: isort::settings::Settings {
+                    required_imports: BTreeSet::from_iter([
+                        // https://github.com/astral-sh/ruff/issues/18729
+                        NameImport::ImportFrom(MemberNameImport::member(
+                            "__future__".to_string(),
+                            "generator_stop".to_string(),
+                        )),
+                        // https://github.com/astral-sh/ruff/issues/16802
+                        NameImport::ImportFrom(MemberNameImport::member(
+                            "collections".to_string(),
+                            "Sequence".to_string(),
+                        )),
+                        // Only bail out if _all_ the names in UP035 are required. `pipes.Template`
+                        // isn't flagged by UP035, so requiring it shouldn't prevent `pipes.quote`
+                        // from getting a diagnostic.
+                        NameImport::ImportFrom(MemberNameImport::member(
+                            "pipes".to_string(),
+                            "Template".to_string(),
+                        )),
+                    ]),
+                    ..Default::default()
+                },
+                ..settings::LinterSettings::for_rules([
+                    Rule::MissingRequiredImport,
+                    Rule::UnnecessaryFutureImport,
+                    Rule::DeprecatedImport,
+                ])
+            },
+        );
+        assert_diagnostics!(diagnostics, @r"
+        UP035 [*] Import from `shlex` instead: `quote`
+         --> <filename>:1:1
+          |
+        1 | from pipes import quote, Template
+          | ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
+          |
+        help: Import from `shlex`
+
+        ℹ Safe fix
+        1   |-from pipes import quote, Template
+          1 |+from pipes import Template
+          2 |+from shlex import quote
+
+        I002 [*] Missing required import: `from __future__ import generator_stop`
+        --> <filename>:1:1
+        help: Insert required import: `from __future__ import generator_stop`
+
+        ℹ Safe fix
+          1 |+from __future__ import generator_stop
+        1 2 | from pipes import quote, Template
+
+        I002 [*] Missing required import: `from collections import Sequence`
+        --> <filename>:1:1
+        help: Insert required import: `from collections import Sequence`
+
+        ℹ Safe fix
+          1 |+from collections import Sequence
+        1 2 | from pipes import quote, Template
+        ");
+    }
 }
